/**********************************************************************************************/
/*  Comments or bug reports to Seungjai Min (seungjai@gmail.com)                              */
/**********************************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "upc_task.h"
#include "timers.h"

/*******************************************************************************/
/* if continuation style execution is needed, uncomment the following line     */
/*******************************************************************************/
#ifdef APPL_DEBUG
#define DEBUG(stmts) stmts
#else
#define DEBUG(stmts)
#endif


/* Checking information */

static int solutions[] = {
    1,
    0,
    0,
    2,
    10, /* N=5 */
    4,
    40,
    92,
    352,
    724, /* N=10 */
    2680,
    14200,
    73712,
    365596,
    2279184,	/* N=15 */
		14772512,
		95815104,
		666090624 /* N=18 */
		/* 4968057848, - too big, need 64-bit int type for result */
		/* 39029188884 - too big, need 64-bit int type for result */
};

#define MAX_SOLUTIONS sizeof(solutions)/sizeof(int)

int total_count;

/* ----------------------------------------------------------------- */

//#define N 16		/* THREADS > 256 */
#define N 14

/* mcutoff: 
		Those taskes that are deeper than this manual cutoff in the execution tree
    will be serialized (executed immediately without pushed into queue 
*/
int mcutoff = (int)(N/2);		/* manual tree-depth cutoff value (default N/2) */

typedef struct data data_t; 
struct data {
	int j;
	int depth;
	char a[N];
	int result[N];
};

/* global task queue variable for task execution */
taskq_handle taskq;

/* timers */
static int UPC_RUN = 0;
static int NQUEENS_TIME = 1;
static int NQUEENS_CONT_TIME = 2;

/*
 * <a> contains array of <n> queen iositions.  Returns 1
 * if none of the queens conflict, and returns 0 otherwise.
 */
int ok(int n, char *a)
{
	int i, j;
	char p, q;

	for (i = 0; i < n; i++) {
		p = a[i];

		for (j = i + 1; j < n; j++) {
			q = a[j];
			if (q == p || q == p - (j - i) || q == p + (j - i))
				return 0;
		}
	}
	return 1;
}

void nqueens_ser (int n, int j, char *a, int *solutions)
{
	int i,res;

	if (n == j) {
		/* good solution, count it */
		*solutions = 1;
		return;
	}

	*solutions = 0;

	/* try each possible position for queen <j> */
	for (i = 0; i < n; i++) {
		/* allocate a temporary array and copy <a> into it */
		a[j] = i;
		if (ok(j + 1, a)) {
			nqueens_ser(n, j + 1, a,&res);
			*solutions += res;
		}
	}
}

void nqueens (data_t *node, int *out)
{
	int i, sum, joinCnt, result[N];
	data_t child;

  DEBUG3(
    fprintf(dbg_fp, "\n----------------------------------------\n");
    fprintf(dbg_fp, "NQUEENS [N=%d j=%d] strt\n", N, node->j);
  )

	if (N == node->j) {
		*out = 1;
		DEBUG3( fprintf(dbg_fp, "1: returning *out=%d\n", *out); )
		return;
	}

	if (node->depth > mcutoff) {
		int ser_result;
		*out = 0;
		for (i = 0; i < N; i++) {
			/* allocate a temporary array and copy <a> into it */
			node->a[node->j] = i;
			if (ok(node->j + 1, node->a)) {
				nqueens_ser(N, node->j + 1, node->a, &ser_result);
				*out += ser_result;
			}
		}

		return;
	}

	joinCnt = 0;
	memset(result, 0, N*sizeof(int));
	for (i = 0; i < N; i++) {
		memcpy(child.a, node->a, N*sizeof(char));
		child.a[node->j] = i;
		if (ok(node->j + 1, child.a)) {
			joinCnt++;
			child.j = (node->j)+1;
			child.depth = (node->depth)+1;
			taskq_put(taskq, nqueens, &child, &result[i]);
		}
	}

	DEBUG3( fprintf(dbg_fp, "joinCnt=%d\n", joinCnt); )

	if (joinCnt > 0) {
		taskq_wait(taskq);
		sum = 0;
		for (i=0; i<N; i++) {
			sum += result[i];
		}	
		*out = sum;
	}
	else {
		*out = 0;
	}
}

void verify_queens()
{
	if ( N > MAX_SOLUTIONS ) printf("SOLUTION not available for N=%d\n", N); 
	if ( total_count == solutions[N-1]) 
		printf("CORRECT: output(%d) == solution(%d)\n", total_count, solutions[N-1]); 
	else
		printf("INCORRECT: output(%d) != solution(%d)\n", total_count, solutions[N-1]); 
}

int main(int argc, char *argv[])
{
	char *a;
	data_t root;
	int steal_size=2;		/* number of tasks to steal */
	int chunk_size=4;		/* number of tasks moved between the private and shared task queue */
	int threshold=8;		/* the size of the private task queue */

	if (argc != 2) {
		printf("USAGE: upcrun -n <nprocs> ./nqueens <mcutoff>\n");
		printf("       (mcutoff: recommended manual cutoff value is slightly less than N/2");
		printf(" i.e. 5~7 for N=12~15)\n"); 
		upc_global_exit(1);
	}

	/* best option: gcc: steal_size=2 chunk_size=4 threshold=8 mcutoff=5 */
	/* best option: icc: steal_size=2 chunk_size=4 threshold=8 mcutoff=7 */
	mcutoff = (int) atoi(argv[1]);

	timer_clear(UPC_RUN);
	timer_clear(NQUEENS_TIME);
	timer_clear(NQUEENS_CONT_TIME);

	taskq = taskq_all_alloc(1, nqueens, sizeof(data_t), sizeof(int));

	taskq_set_steal_size(taskq, steal_size);
	taskq_set_chunk_size(taskq, chunk_size);
	taskq_set_threshold(taskq, threshold);

	upc_barrier;

	total_count=0;

	if (MYTHREAD==0) {
		root.j = 0;
		root.depth = 0;
		memset(root.result, 0, N*sizeof(int));
		taskq_put(taskq, nqueens, &root, &total_count);
	}

	timer_start(UPC_RUN);

	taskq_all_run(taskq);

	timer_stop(UPC_RUN);

	upc_barrier;

	if (MYTHREAD==0) printf(" completed!\n");

	taskq_check(taskq);
	taskq_stats(taskq);

	if (MYTHREAD==0) {
		verify_queens();
		printf("steal_size=%d, chunk_size=%d, threshold=%d\n", steal_size, chunk_size, threshold);
		printf("UPC_RUN=%lf sec\n", timer_read(UPC_RUN));
	}

	return 0;
}
