/*
 *         ---- The Unbalanced Tree Search (UTS) Benchmark ----
 *  
 *  Copyright (c) 2010 See AUTHORS file for copyright holders
 *
 *  This file is part of the unbalanced tree search benchmark.  This
 *  project is licensed under the MIT Open Source license.  See the LICENSE
 *  file for copyright and licensing information.
 *
 *  UTS is a collaborative project between researchers at the University of
 *  Maryland, the University of North Carolina at Chapel Hill, and the Ohio
 *  State University.  See AUTHORS file for more information.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "uts.h"

/**** UPC Definitions ****/
#include <upc.h>
#include "upc_task.h"
#define SHARED_INDEF     shared [0]
#define VOLATILE         strict

//#define MANUAL_CUTOFF
//#define TASK_TIMER  // timer for measuring task function exec. time

#ifdef TASK_TIMER
int counter = 0;
double task_time = 0.0;
#endif

int steal_size, chunk_size, threshold;
int chunk_mode;

/***********************************************************
 *  Parallel execution parameters                          *
 ***********************************************************/

int doSteal   = 1; 				// 1 => use work stealing
int cbint     = 1;        // Cancellable barrier polling interval
int pollint   = 1;        // BUPC Polling interval

shared int nNodes[THREADS];
shared int nLeaves[THREADS];
int my_nNodes = 0;
int my_nLeaves = 0;
int max_height = 0;
int cutoff=1073741824;
int rcut_coeff = 20;

void treeSearch(Node *);

typedef struct data data_t;
struct data {
	Node parent;
	Node child;
};

/***********************************************************
 * Tree statistics (if selected via UTS_STAT)              *
 *   compute overall size and imbalance metrics            *
 *   and histogram size and imbalance per level            *
 ***********************************************************/
int stats   = 0;
int unbType = -1;

/***********************************************************
 *  Global shared state                                    *
 ***********************************************************/
taskq_t *taskq;

shared double startTime[THREADS];

/***********************************************************
 *  UTS Implementation Hooks                               *
 ***********************************************************/

// Return a string describing this implementation
char * impl_getName() {
  char * name = "UPC";
  return name;
}

// construct string with all parameter settings 
int impl_paramsToStr(char *strBuf, int ind) {
  ind += sprintf(strBuf+ind, "Execution strategy:  ");
  ind += sprintf(strBuf+ind, "Parallel search using %d threads\n", THREADS);
  if (doSteal) {
    ind += sprintf(strBuf+ind, "  Load balance by work stealing\n");
    ind += sprintf(strBuf+ind, "  CBarrier Interval: %d\n", cbint);
    ind += sprintf(strBuf+ind, "  Polling Interval: %d\n", pollint);
  }
  else
    ind += sprintf(strBuf+ind, "   No load balancing.\n");
      
  return ind;
}

int impl_parseParam(char *param, char *value) {
  int err = 0;  // Return 0 on a match, nonzero on an error

  switch (param[1]) {
		case 'x':
			cutoff = atoi(value); break;
		case 'y':
			rcut_coeff = atoi(value); break;
    case 'l':
      steal_size = atoi(value); break;
    case 'c':
      chunk_size = atoi(value); break;
    case 'o':
      threshold = atoi(value); break;
    case 'k':
      chunk_mode = atoi(value); break;
    case 'i':
      cbint = atoi(value); break;
#ifdef __BERKELEY_UPC__
    case 'I':
      pollint = atoi(value); break;
#endif
    default:
      err = 1;
      break;
  }

  return err;
}

void impl_helpMessage() {
  printf("   -s  int   zero/nonzero to disable/enable work stealing\n");
  printf("   -c  int   chunk size for work stealing\n");
  printf("   -k  int   chunk mode for work stealing (1:FIXED, 2:HALF, 3:ADAPTIVE)\n");
  printf("   -i  int   set cancellable barrier polling interval\n");
#ifdef __BERKELEY_UPC__
  printf("   -I  int   set working bupc_poll() interval\n");
#endif
}

void impl_abort(int err) {
  upc_global_exit(err);
}


/***********************************************************
 *                                                         *
 *  FUNCTIONS                                              *
 *                                                         *
 ***********************************************************/

/*
 *	Tree Implementation      
 *
 */
void initNode(Node * child)
{
  child->type = -1;
  child->height = -1;
  child->numChildren = -1;    // not yet determined
}


void initRootNode(Node * root, int type)
{
  uts_initRoot(root, type);
}


/* 
 * Generate all children of the parent
 *
 * details depend on tree type, node type and shape function
 *
 */
void treeSearch(Node *parent) 
{
  int parentHeight = parent->height;
  int numChildren, childType;
	Node child;
#ifdef TASK_TIMER
	double t1, t2;
  //t1 = uts_wctime();
#endif

	my_nNodes++; 

  numChildren = uts_numChildren(parent);
  childType   = uts_childType(parent);

  // record number of children in parent
  parent->numChildren = numChildren;
  
  // construct children and push onto stack
  if (numChildren > 0) {
    int i, j;
    child.type = childType;
    child.height = parentHeight + 1;

		if (max_height < child.height) max_height = child.height;

    for (i = 0; i < numChildren; i++) {
      for (j = 0; j < computeGranularity; j++) {
        // TBD:  add parent height to spawn
        // computeGranularity controls number of rng_spawn calls per node
#ifdef TASK_TIMER
  			t1 = uts_wctime();
#endif
        rng_spawn(parent->state.state, child.state.state, i);
#ifdef TASK_TIMER
				t2 = uts_wctime();
				task_time += (t2 - t1); 	
				counter++;
#endif
      }

#ifdef MANUAL_CUTOFF
			if (child.height > cutoff) {
				treeSearch(&child);
			}
			else
      	taskq_put(taskq, treeSearch, &child);
#else
      	taskq_put(taskq, treeSearch, &child);
#endif

    }
  } 
	else {
    my_nLeaves++;
  }

}

// display search statistics
void showStats(double elapsedSecs) {
  int i;
  int tnodes = 0, tleaves = 0, trel = 0, tacq = 0, tsteal = 0, tfail= 0;
  int mdepth = 0, mheight = 0, tserial = 0;
  double twork = 0.0, tsearch = 0.0, tidle = 0.0, tovh = 0.0, tcbovh = 0.0;
	shared rtaskq_t *rtaskq_sp;

  // combine measurements from all threads
  for (i = 0; i < THREADS; i++) {
		rtaskq_sp = (shared rtaskq_t *) &(taskq->rtaskq_sp[i]);
		//printf("pid=%d: nNodes = %d nSteal = %d nTaskSteal = %d\n", i, nNodes[i], 
		//				rtaskq_sp->nSteal, rtaskq_sp->nTaskSteal);
    tnodes  += nNodes[i];
    tleaves += nLeaves[i];
    trel    += rtaskq_sp->nRelease;
    tacq    += rtaskq_sp->nAcquire;
    tsteal  += rtaskq_sp->nTaskSteal;
    tfail   += rtaskq_sp->nFail;
  }
  if (trel != tacq + tsteal) {
    printf("*** error! total released != total acquired + total stolen\n");
  }
    
  uts_showStats(THREADS, steal_size, elapsedSecs, tnodes, tleaves, mheight);

  if (verbose > 1) {
    if (doSteal) {
      printf("Total chunks released = %d, of which %d reacquired and %d stolen\n",
          trel, tacq, tsteal);
      printf("Failed steal operations = %d, ", tfail);
    }
  }
}


/*  Main() function
 *  Notes on execution model:
 *     - under openMP, global vars are all shared
 *     - under UPC, global vars are private unless explicitly shared
 *     - UPC is SPMD starting with main, OpenMP goes SPMD after
 *       parsing parameters
 */
int main(int argc, char *argv[]) {
	data_t data;
  Node root;

	// USAGE: upcrun -n 8 uts $T1L -l <steal_size> -c <chunk_size> -o <threshold> -k <chunk_mode=1(FIXED),2(HALF),3(ADAPTIVE)>
	// best configuration T1L: s4c4t4  T2L: s4c4t4

  /* determine benchmark parameters (all PEs) */
  uts_parseParams(argc, argv);

  {
    double t1, t2, et;

    /* show parameter settings */
    if (MYTHREAD == 0) {
			//printf("[USAGE] upcrun -n 8 uts $T1L -l <steal-size> -c <chunk_size> -o <threshold> -k <chunk_mode:1(FIXED),2(HALF),3(HIER)>\n");
      uts_printParams();
    }

    /* initialize stealstacks */
		taskq = taskq_all_alloc(1, treeSearch, sizeof(Node), 0);

		steal_size = 10;
		chunk_size = 20;
		threshold = 20;
		chunk_mode = 1;
		taskq_set_steal_size(taskq, steal_size);
		taskq_set_chunk_size(taskq, chunk_size);
		taskq_set_threshold(taskq, threshold);
		taskq_set_independent_task(taskq, treeSearch);
		taskq_set_runtime_cutoff(taskq, rcut_coeff);
		taskq_set_chunk_mode(taskq, chunk_mode);

/*  RANDOM: SMP */
		taskq_set_steal_mode(taskq, FULL_RAND_STEAL, 1, THREADS);

    /* initialize root node and push on thread 0 stack */
    if (MYTHREAD == 0) {
			/* type is initialized to "GEO" in uts_common.c and is set using "-t" in uts_parseParams func */
      initRootNode(&root, tree_type );
      taskq_put(taskq, treeSearch, &root);
    }

    // line up for the start
/* #pragma omp barrier */
    upc_barrier;
    
    /* time parallel search */
    t1 = uts_wctime();

    taskq_all_run(taskq);

    t2 = uts_wctime();
    et = t2 - t1;

/* #pragma omp barrier */
		nNodes[MYTHREAD] = my_nNodes;
		nLeaves[MYTHREAD] = my_nLeaves;

    upc_barrier;

		taskq_stats(taskq);
		taskq_check(taskq);

    /* display results */
    if (MYTHREAD == 0) {
			printf("max_height=%d\n", max_height);
      showStats(et);
    } 

#ifdef TASK_TIMER
		fprintf(dbg_fp, "p%d: task func count=%d total time=%f avg time=%f\n", MYTHREAD, 
						counter, task_time, (double)task_time*1000000000/(double)counter);
#endif
  }
/********** End Parallel Region **********/

  return 0;
}
