#include "upc_task.h"

/*
#define DBG_MAP_LEVEL1
#define DBG_MAP_LEVEL2
#define DBG_MAP_LEVEL3
*/

#ifdef DBG_MAP_LEVEL1
#define DBG_MAP1(stmts) stmts 
#else
#define DBG_MAP1(stmts)
#endif

#ifdef DBG_MAP_LEVEL2
#define DBG_MAP2(stmts) stmts 
#else
#define DBG_MAP2(stmts)
#endif

#ifdef DBG_MAP_LEVEL3
#define DBG_MAP3(stmts) stmts 
#else
#define DBG_MAP3(stmts)
#endif

/***********************************************************
 *                Hash Map in plain C                      *
 ***********************************************************/

inline uint64_t hash_uint64(uint64_t inKey)
{
  // Thomas Wang's 64 bit integer hash.
	uint64_t hash = (uint64_t)inKey;
	hash += ~(hash << 32);
	hash ^=  (hash >> 22);
	hash += ~(hash << 13);
	hash ^=  (hash >> 8);
	hash +=  (hash << 3);
	hash ^=  (hash >> 15);
	hash += ~(hash << 27);
	hash ^=  (hash >> 31);
	return hash;
}

void print_free_entries(upc_map_t *map) {
	entry_t *entry;
	fprintf(dbg_fp, "\nfree_entries list: ");
	for (entry=map->free_entries; entry!=NULL; entry=entry->next_free) { 
		fprintf(dbg_fp, "%p ", entry); 
	}
	fprintf(dbg_fp, "\n\n");

	return;
}

/* 
	map structure = bucket array + entries array + values array
  buckets: there are num_buckets buckets where each element is an (entry_t *) type, 
          but it is used as a place holder for a pointer to the actual entry list
  entries: a singly linked list whose size is pre-allocated to num_entries
  values: each entry has a matching value element in values array, i.e., the entry 
					stored in entries[10] has a value stored in values[10]
 */

upc_map_t *upc_map_alloc(const char *name, int num_buckets, int num_entries, int val_size)
{
	upc_map_t *map;
	entry_t *entry, *bucket;

	DBG_MAP1( fprintf(dbg_fp, "upc_map_alloc(num_buckets=%d, num_entries=%d, val_size=%d)\n", 
					num_buckets, num_entries, val_size); )

	map = (upc_map_t *) malloc( sizeof(upc_map_t) );

	strcpy(map->name, name);
	map->num_buckets = num_buckets;
	map->num_entries = num_entries;
	map->val_size    = val_size;
	map->entry_cnt   = 0;
	map->buckets     = (entry_t *) malloc(num_buckets * sizeof(entry_t));
	map->entries     = (entry_t *) malloc(num_entries * sizeof(entry_t));
	map->values      = (char *) malloc(num_entries * val_size);

	/* initially, the entries and the free_entries are the same */
	map->free_entries = (entry_t *)map->entries;	

	for (int i=0; i<map->num_buckets; i++) {
		bucket = (entry_t *) &map->buckets[i];
		bucket->next = NULL;
		// DBG_MAP1( fprintf(dbg_fp, "  &map->buckets[%d] = %lld\n", i, (char *) &map->buckets[i]); )
	}

	for (int i=0; i<map->num_entries; i++) {
		entry = (entry_t *) &map->entries[i];
		entry->next = NULL;
		entry->next_free = (entry_t *) &map->entries[i+1];
		if (i == map->num_entries-1 ) entry->next_free = NULL;
		// DBG_MAP1 ( fprintf(dbg_fp, "  &map->entries[%d] = %lld\n", i, entry); )
	}

	DBG_MAP1 (
		fprintf( dbg_fp, "  [upc_map_alloc] allocated %s map at %p\n", map->name, map);
		fprintf( dbg_fp, "  [upc_map_alloc] allocated map->buckets[%d : %d] at [%p : %p]\n", 
												0, num_buckets-1, &map->buckets[0], &map->buckets[num_buckets-1]);
		fprintf( dbg_fp, "  [upc_map_alloc] allocated map->entries[%d : %d] at [%p : %p]\n", 
												0, num_entries-1, &map->entries[0], &map->entries[num_entries-1]);
		fprintf( dbg_fp, "  [upc_map_alloc] allocated map->values[%d : %d] at [%p : %p]\n", 
												0, num_entries-1, &map->values[0], map->values+(num_entries-1)*val_size);
	)
	
	/* make sure you this barrier ensures that memory is all allocated before the application starts */
	upc_barrier;	/* do not remove this barrier */

	return map;
}

/* this map does not check if the same key has been already inserted before*/
void *upc_map_put(upc_map_t *map, uint64_t key, void *value)
{ 
	uint64_t hashID;
	entry_t *last_entry;
	entry_t *new_entry;
	int new_entry_index;

	DBG_MAP2( 
		fprintf(dbg_fp, "  [upc_map_put] %s: map=%p key=%lld value=%lld\n", map->name, map, key, *(uint64_t *)value);
	)

	if (map->entry_cnt >= map->num_entries) {
		printf("  [upc_map_put] Map %s Overflow: tid=%d, entry_cnt=%d >= num_entries=%d\n", 
						map->name, MYTHREAD, map->entry_cnt, map->num_entries);
		upc_global_exit(1);
	}

	DBG_MAP3(
		if (map->free_entries==NULL) { 
			printf("  [upc_map_put] %s tid%d: map->free_entries is NULL\n", map->name, MYTHREAD); 
			upc_global_exit(1);
		}
	)

	hashID = hash_uint64(key) % map->num_buckets;

	DBG_MAP3 (
		fprintf(dbg_fp, "  [upc_map_put] tid%d: key=%lld hashID=%lld ", MYTHREAD, key, hashID);
		fprintf(dbg_fp, " entry_cnt=%d free_entries=%p\n", map->entry_cnt, map->free_entries);
	)

	last_entry = (entry_t *) &map->buckets[hashID];
	DBG_MAP1( fprintf(dbg_fp, "  [upc_map_put] last_entry0=%p\n", last_entry); )
	while (last_entry->next != NULL) {
		last_entry = last_entry->next;
		DBG_MAP1( fprintf(dbg_fp, "  [upc_map_put] last_entry=%p\n", last_entry); )
	}

	new_entry = map->free_entries;

	DBG_MAP3( fprintf(dbg_fp, "  [upc_map_put] map->free_entries=%p\n", map->free_entries); )

	last_entry->next = new_entry;
	map->free_entries = map->free_entries->next_free;

	DBG_MAP3( fprintf(dbg_fp, "  [upc_map_put] NEW map->free_entries=%p\n", map->free_entries); )

	new_entry->key = key;
	new_entry_index = ( new_entry - map->entries );		/* offset from the start address */

	DBG_MAP3( fprintf(dbg_fp, "  [upc_map_put] new_entry=%p map->entries=%p, new_entry_index=%d\n", 
				(char *)new_entry, (char *)map->entries, new_entry_index); )

	new_entry->val = (void *) (map->values + map->val_size*new_entry_index);

	DBG_MAP2(
		fprintf(dbg_fp, "  [upc_map_put] memcpy(%p, %p, %d) ", new_entry->val, value, map->val_size);
		fprintf(dbg_fp, "*value = %lld\n", *(uint64_t *)value);
	)

	memcpy(new_entry->val, value, map->val_size);
	new_entry->next = NULL;
	map->entry_cnt++;

	return new_entry->val;
}

void *upc_map_remove(upc_map_t *map, uint64_t key) 
{
	uint64_t hashID = hash_uint64(key) % map->num_buckets;
	int key_found = 0;
	entry_t *prev_entry = (entry_t *) &map->buckets[hashID]; 
	entry_t *curr_entry = prev_entry->next;
	entry_t *prev_free_entry_head;

	DBG_MAP2 ( fprintf(dbg_fp, "  [upc_map_remove] %s: key=%lld hashID=%lld entry_cnt=%d\n", 
					map->name, key, hashID, map->entry_cnt);
	)

	if (map->entry_cnt == 0) return NULL; 

	while (curr_entry != NULL) {
		if (curr_entry->key == key) { key_found=1; break; }
		prev_entry = curr_entry;
		curr_entry = curr_entry->next;
	}

	if (key_found) {		/* entry found! */
		prev_entry->next = curr_entry->next;
		curr_entry->next = NULL;
		map->entry_cnt--;

		prev_free_entry_head = map->free_entries;
		map->free_entries = curr_entry;
		map->free_entries->next_free = prev_free_entry_head;
	}
	else { return NULL; }

	DBG_MAP2( fprintf(dbg_fp, "  [upc_map_remove] map(%s) successfully removed key %lld\n", map->name, key); )
	return (void *)curr_entry->val;
}

void *upc_map_get(upc_map_t *map, uint64_t key) 
{
/*
	uint64_t hashID = hash_uint64(key) % map->num_buckets;
	entry_t *prev_entry = (entry_t *) &map->buckets[hashID]; 
	entry_t *curr_entry = prev_entry->next;
*/
	uint64_t hashID;
	entry_t *prev_entry;
	entry_t *curr_entry;

	hashID = hash_uint64(key) % map->num_buckets;
	prev_entry = (entry_t *) &map->buckets[hashID]; 
	curr_entry = prev_entry->next;

	DBG_MAP2 (
		fprintf(dbg_fp, "  [upc_map_get] %s: map=%p key=%lld hashID=%lld entry_cnt=%d\n", map->name, map, key, hashID, map->entry_cnt);
	)

	if (map->entry_cnt == 0) return NULL; 

	while (curr_entry != NULL) {
		DBG_MAP3( fprintf(dbg_fp, "  curr_entry: key=%lld val=%lld\n", curr_entry->key, 
							*(uint64_t *)curr_entry->val); )
		if (curr_entry->key == key) { 
			DBG_MAP3( fprintf(dbg_fp, "  found and returning val=%lld\n", *(uint64_t *)curr_entry->val); )
			DBG_MAP3( fprintf(dbg_fp, "  map=%s: key=%lld val=%lld\n", map->name, curr_entry->key, 
								*(uint64_t *)curr_entry->val); )
			return (void *)curr_entry->val; 
		}
		prev_entry = curr_entry;
		curr_entry = curr_entry->next;
	}

	DBG_MAP2( fprintf(dbg_fp, "  [upc_map_get] returning NULL\n"); )
	return NULL;
}

void upc_map_free(upc_map_t *map)
{
	free(map->buckets);
	free(map->entries);
	free(map->values);
	free(map);
}

int upc_map_size(upc_map_t *map)
{
	return map->num_buckets;
}

void upc_map_stats(upc_map_t *map) { upc_map_snapshot(map, 0); }
void upc_map_display(upc_map_t *map) { upc_map_snapshot(map, 1); }

void upc_map_snapshot(upc_map_t *map, int flag)
{
	int i, j, count;
	entry_t *entry;

	DBG_MAP2( fprintf(dbg_fp, "  [upc_map_display]\n"); )
	fprintf(dbg_fp, "\n--------------- Hashmap snapshot ---------------\n");

	for (i=0; i<map->num_buckets; i++) {
		count=0;
		entry = (entry_t *) &map->buckets[i];
		fprintf(dbg_fp, "entry[%d]: ", i);
		entry = entry->next;
		while (entry != NULL) {
			if (flag) fprintf(dbg_fp, " [%lu, %d]", entry->key, *(int *)entry->val);
			entry = entry->next;
			count++;
		}
		fprintf(dbg_fp, " count=%d\n", count);
	}
}


void *upc_map_get_one(upc_map_t *map)
{
	int i, j, count;
	entry_t *entry;

	DBG_MAP2( fprintf(dbg_fp, "  [upc_map_display]\n"); )
	fprintf(dbg_fp, "\n--------------- Hashmap snapshot ---------------\n");

	for (i=0; i<map->num_buckets; i++) {
		count=0;
		entry = (entry_t *) &map->buckets[i];
		fprintf(dbg_fp, "entry[%d]: ", i);
		entry = entry->next;
		while (entry != NULL) {
			fprintf(dbg_fp, " [%lu, %d]\n", entry->key, *(int *)entry->val);
			return (void *)entry->val; 
		}
		fprintf(dbg_fp, " count=%d\n", count);
	}

	return NULL;
}
